.cell.baobab-level-cell,
.cell.baobab-level-cell:hover,
.cell.baobab-level-cell:selected,
.cell.baobab-level-cell:selected:hover {
    border-color: darker(@borders);
    border-width: 1px;
    border-radius: 3px;
    border-style: solid;
    padding: 1px;
    background-color: white;
}

.cell.baobab-level-cell.fill-block,
.cell.baobab-level-cell.fill-block:selected,
.cell.baobab-level-cell.fill-block:selected:hover {
    background-color: #edd400;
}

.cell.baobab-level-cell.fill-block.level-low,
.cell.baobab-level-cell.fill-block.level-low:hover {
    background-color: #73d216;
}

.cell.baobab-level-cell.fill-block.level-high,
.cell.baobab-level-cell.fill-block.level-high:hover {
    background-color: #cc0000;
}

.cell.baobab-level-cell.fill-block:backdrop,
.cell.baobab-level-cell.fill-block:hover:backdrop,
.cell.baobab-level-cell.fill-block.level-low:backdrop,
.cell.baobab-level-cell.fill-block.level-high:backdrop {
    background-color: @theme_unfocused_text_color;
}

.cell.baobab-cell-error {
    color: @error_color;
}

.cell.baobab-cell-warning {
    color: @warning_color;
}

.cell.baobab-cell-warning:selected,
.cell.baobab-cell-error:selected {
    color: @theme_selected_fg_color;
}

.cell.baobab-cell-warning:backdrop,
.cell.baobab-cell-error:backdrop,
.cell.baobab-cell-warning:selected:backdrop,
.cell.baobab-cell-error:selected:backdrop {
    color: @theme_unfocused_text_color;
}

window.background stack > grid > scrolledwindow.frame {
    border-radius: 0;
    border-width: 0 1px 0 0;
}

window infobar.warning,
window infobar.error {
    border-bottom-width: 1px;
    border-bottom-style: solid;
    border-bottom-color: @borders;
}

ringschart {
    padding: 13px 13px 13px 13px;
}

ringschart.subfolder-tip {
    color: white;
    background-color: rgba(0, 0, 0, 0.8);
    border-radius: 3px;
    padding: 3px 3px 3px 3px;
}

